/* Sfcn_UDPdrv.C *****************************************************
 *
 */

/* Observaes ***************************************************
 *
 */

/* Assuntos diversos *********************************************/

/* Macros Simulink */
#define S_FUNCTION_NAME  Sfcn_UDPdrv
#define S_FUNCTION_LEVEL 2


/* Macros Gerais */
#include <stdio.h>
#include "winsock2.h"
#include "simstruc.h"
#include "time.h"
#include "mex.h"
#include "TSCaccess.c"

#define word unsigned short int 
#define dword unsigned int
#define PORT 11111  // porta do equipamento Lynx

#include "lynx_ac1160.h"

#define PARAM_STRLEN (32) 

// Parmetros
char_T   IPAddrStr[PARAM_STRLEN];   // End. IP passado pelo simulink
double   WorkFreq;                  // Frequencia de aquisiao
double   CalFreq;                   // Frequencia calibrada do TSC
unsigned char RangeIdx;             // Range de operao
double   BaseFreq;                  // Frequencia base do mdulo Lynx
unsigned char ganho;                // ganho das entradas analogicas
double escala;                      // range das entradas analogicas

// UDP
WSADATA wsaData;                    //biblioteca UDP
SOCKET ListenSocket;                //Socket para receber pacotes 
SOCKADDR_IN addr_send;              // End. Socket p/ enviar pacotes
SOCKADDR_IN addr_recv;              // End. Socket p/ receber pacotes
int addr_recv_len = sizeof(addr_recv);//tamanho da estrutura addr_recv
int addr_send_len = sizeof(addr_send);//tamanho da estrutura addr_send
int errodados;                       // indica erro de pacotes recebidos  
u_long nonblockingmode = 1; //define modo nonblocking (l pacotes instantaneamente) 

// declarados aqui devido a um bug do Visual C
//Estruturas de comandos
TpPcInfoReq       informa; // p/ pedir informaes sobre o equipamento
TpPcGetInfoR      rinfo; // resposta do pedido de informao
TpPcAcqSetup      setup; // configurar equipamento
TpPcAnswerR       rsetup; // resposta do setup
TpPcInfoReq       start; // Inicia aquisio de dados (apenas leitura)
TpPcAnswerR       rstart;// resposta do start
TpPcInfoReq       stop; // parar aquisio de dados
TpPcAnswerR       rstop;// resposta do stop
TpPcAcqData       dados;// pacotes com as amostras
TpPcAdsWriteDO    dout;// escrita em saida digital
TpPcAdsWriteDO    aout;// escrita em saida analogica
TpPcAdsWriteDO_R  rdout;// resposta de escrita em saida

// temporizao
ULONGLONG basetscfirst, basetsclast; // base de temporizao
double  remainingtime;// tempo restante em cada Scan 
BOOL IsFirstStep; // flag p/ primeiro loop de Scan

/* Rotinas Simulink **********************************************/

/* Inicializao */
static void mdlInitializeSizes(SimStruct *S){
   int i, j; // contadores
   int buffsize = 66*0; // tamanho buffer de recepo
   errodados = 0; // reinicia contagem de erro de pacote
   IsFirstStep = TRUE; // inicia no primeiro loop
   ssSetNumSFcnParams(S,5);// N de parametros do bloco
   if (ssGetNumSFcnParams(S) != ssGetSFcnParamsCount(S)) return;
   if (!ssSetNumOutputPorts(S,3)) return;// N portas de saida do bloco.
   ssSetOutputPortWidth(S,0,16);// tamanho da saida "0" - entradas analogicas do equipamento
   ssSetOutputPortWidth(S,1,16);// tamanho da saida "1" - entradas digitais do equipamento.
   ssSetOutputPortWidth(S,2,3);// tamanho da saida "2" - status
   if (!ssSetNumInputPorts(S,2)) return;// N portas de entrada do bloco.
   ssSetInputPortWidth(S,0,16);// tamanho da entrada "0" - saidas analogicas do equipamento
   ssSetInputPortWidth(S,1,16);// tamanho da entrada "1" - saidas digitais do equipamento
   ssSetInputPortDirectFeedThrough(S,0,1);// permite passagem direta entre entrada e saida
   ssSetInputPortDirectFeedThrough(S,1,1);
   ssSetNumContStates(S,0);// "0" Estados continuos
   ssSetNumDiscStates(S,0);// "0" Estados discretos
   ssSetNumSampleTimes(S, 1);// apenas 1 taxa de amostragem
   ssSetOptions(S, SS_OPTION_EXCEPTION_FREE_CODE | SS_OPTION_CALL_TERMINATE_ON_EXIT);
   // em caso de erro vai p/ mdlTerminate
    
   // recebimento dos parmetros
   mxGetString(ssGetSFcnParam(S,0), IPAddrStr, PARAM_STRLEN); // end. IP
   WorkFreq  = mxGetScalar(ssGetSFcnParam(S,1));//Freq de Trabalho
   CalFreq  = mxGetScalar(ssGetSFcnParam(S,2));// Calibrao de Frequencia
   RangeIdx = mxGetScalar(ssGetSFcnParam(S,3));//Range das entradas
   BaseFreq = mxGetScalar(ssGetSFcnParam(S,4));// Frequencia de envido de pacotes do equipamento- default:4kHz
   
   // Proteo dos valores dos parametros do bloco
   if (BaseFreq < WorkFreq){
     ssSetErrorStatus(S, "\nUDPdrv: Freqncia base do mdulo menor que a freqncia de trabalho");
     return;
   }  
   
   if (BaseFreq < 1){
     ssSetErrorStatus(S, "\nUDPdrv: Freqncia base do mdulo menor que 1 Hz");
     return;
   }  
   
   if (WorkFreq < 1){
     ssSetErrorStatus(S, "\nUDPdrv: Freqncia de trabalho menor que 1 Hz");
     return;
   }
   // estabelece conexo com a biblioteca WinSock2
   if (WSAStartup(MAKEWORD(2,2), &wsaData) != 0)
     printf("\nUDPdrv: Impossvel estabelecer conexo");
   // Abre novo Socket  
   if ((ListenSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == INVALID_SOCKET){
     WSACleanup();
     printf("\nUDPdrv: Impossvel criar socket");
   }

   //Define tamanho do buffer de recepo de pacotes
   if (setsockopt(ListenSocket,SOL_SOCKET,SO_RCVBUF,&buffsize,sizeof(buffsize)) != 0)
     printf("\nUDPdrv: Impossvel configurar o socket");

   // Ajusta endereo IP e Porta para recepo de pacotes
   addr_recv.sin_family = AF_INET;
   addr_recv.sin_addr.s_addr = htonl(INADDR_ANY);
   addr_recv.sin_port = htons(PORT);

   // Associa Socket  Porta, End. IP p/ recepo de pacotes
   if (bind(ListenSocket,(SOCKADDR*) &addr_recv, addr_recv_len) == SOCKET_ERROR){
     closesocket(ListenSocket);
     printf("\nUDPdrv: Impossvel fazer o bind do socket");
   }

   // Ajusta endereo IP e Porta para envio dos pacotes
   addr_send.sin_family = AF_INET;
   addr_send.sin_port = htons(PORT);
   addr_send.sin_addr.s_addr = inet_addr(IPAddrStr);

   // Configura modo nonblocking
   ioctlsocket(ListenSocket, FIONBIO, &nonblockingmode);
   
   // Configura parmetros do AC1160
   setup.PacketCode = pcAcqSetup;
   setup.Reserved = 0x0000;
   setup.TrigMode = 0x0000; // no implementado
   setup.PreScaler = 1;
   setup.ClockDiv =  (4000.0/BaseFreq - 1.0);// calculo p/ definir BaseFreq
   setup.Mode = 0x0000;// configurado conforme Lynx_ac1160.h
   setup.SampleFreq = 0;// no implementado
   setup.nBurstPack = 1;// para cada aquisio envia um pacote
   setup.nSignals = 17;// 16 entradas analogicas + 1 digital (16 bits)
   // seleciona ganho e range das entradas analogicas (16 bits)
   switch( RangeIdx ){
     case 1: //-1 a 1 V
       ganho = 6;
       escala = 2.0/65536;
       break;
     case 2:  // -2 a 2 V
       ganho = 4;
       escala = 4.0/65536;
       break;
     case 3: // -5 a 5 V
       ganho = 2;
       escala = 10.0/65536;
       break;
     case 4: // -10 a 10 V
       ganho = 0;
       escala = 20.0/65536;
       break;
   }
   // configurao das entradas conforme Lynx_ac1160.h
   setup.ac_icm[0] = 0x8000+ganho;// config entrada analogica "0"
   setup.ac_icm[1] = 0x8008+ganho;// config entrada analogica "1"
   setup.ac_icm[2] = 0x8010+ganho;// config entrada analogica "2"
   setup.ac_icm[3] = 0x8018+ganho;// config entrada analogica "3"
   setup.ac_icm[4] = 0x8020+ganho;// config entrada analogica "4"
   setup.ac_icm[5] = 0x8028+ganho;// config entrada analogica "5"
   setup.ac_icm[6] = 0x8030+ganho;// config entrada analogica "6"
   setup.ac_icm[7] = 0x8038+ganho;// config entrada analogica "7"
   setup.ac_icm[8] = 0x8040+ganho;// config entrada analogica "8"
   setup.ac_icm[9] = 0x8048+ganho;// config entrada analogica "9"
   setup.ac_icm[10] = 0x8050+ganho;// config entrada analogica "10"
   setup.ac_icm[11] = 0x8058+ganho;// config entrada analogica "11"
   setup.ac_icm[12] = 0x8060+ganho;// config entrada analogica "12"
   setup.ac_icm[13] = 0x8068+ganho;// config entrada analogica "13"
   setup.ac_icm[14] = 0x8070+ganho;// config entrada analogica "14"
   setup.ac_icm[15] = 0x8078+ganho;// config entrada analogica "15"
   setup.ac_icm[16] = 0x0000;// config entrada digital (16 bits)

   // envia comando de Setup
   if (sendto(ListenSocket, (char *)&setup, sizeof(setup), 0,(SOCKADDR *)&addr_send, addr_send_len) == -1)
     printf( "\nUDPdrv: Impossvel executar a configurao do dispositivo!");
   
   // aguarda resposta do Setup 
   Sleep (200);
   // l resposta do Setup
   if (recvfrom(ListenSocket, (char *)&rsetup, sizeof(rsetup), 0,(SOCKADDR *)&addr_recv, &addr_recv_len) == -1)
     printf("\nUDPdrv: Sem resposta ao setup do dispositivo!");
  
   // verificao de erro no Setup
   if (rsetup.PacketCode == pcAcqSetupR && rsetup.Error == 0)
     printf("\nUDPdrv: Configuraes OK");
   else
     printf("\nUDPdrv: Erro nas Configuraes");
      
   //Inicia em 0 V as 16 sadas analgicas
   aout.PacketCode = pcWriteDO;
   aout.Reserved = 0;
   aout.AddrADS = 0; 
   aout.Value = 0x0000;
   for (i=0;i<16;i++){
     aout.AddrADS = (16+i) << 8; // end. dos canais analogicos comeando em 16 no Byte + significativo 
     // envia comando de escrita na saida
     if (sendto(ListenSocket, (char *)&aout, sizeof(aout), 0,(SOCKADDR *)&addr_send, addr_send_len) == -1)
        printf("\nUDPdrv: Impossvel gravar valores na sada analgica(%d)", i+1);
     waitbusy(0.4e-3, gettsc(), CalFreq);// tempo de espera p/ equipamento reconhecer comandos de escrita
   }

   //Inicia em "0" as sadas digitais
   dout.PacketCode = pcWriteDO;
   dout.Reserved = 0x0000;
   dout.AddrADS = 0x0000;
   dout.Value = 0x0000;
   // envia comando de escrita na saida digital
   if (sendto(ListenSocket, (char *)&dout, sizeof(dout), 0,(SOCKADDR *)&addr_send, addr_send_len) == -1)
     printf("\nUDPdrv: Impossvel gravar valores na sada digital");
   waitbusy(1e-3, gettsc(), CalFreq);// tempo de espera p/ equipamento reconhecer comandos de escrita
   // Consome pacote de reconhecimento de escrita
   if (recvfrom(ListenSocket, (char *)&rstart, sizeof(rstart), 0,(SOCKADDR *)&addr_recv, &addr_recv_len) == -1)
     printf("\nUDPdrv: Erro na escrita na sada digital", i+1);
   
   // Inicia Aquisies 
   start.PacketCode = pcAcqStart;
   start.Request = 0x0000;
   start.Reserved = 0x0000;
   // envia comando para iniciar aquisies
   if (sendto(ListenSocket, (char *)&start, sizeof(start), 0,(SOCKADDR *)&addr_send, addr_send_len) == -1)
     printf("\nUDPdrv: Impossvel iniciar as aquisies!");
   
   basetscfirst = gettsc(); //contagem de tempo  partir do inicio do envio de pacotes 

   waitbusy(10e-3, gettsc(), CalFreq);// tempo de espera p/ equipamento reconhecer comandos 
   
   // Resposta do comando de iniciar aquisio
   if (recvfrom(ListenSocket, (char *)&rstart, sizeof(rstart), 0,(SOCKADDR *)&addr_recv, &addr_recv_len) == -1)
     printf("\nUDPdrv: Sem resposta do dispositivo para inicializao!");
   
   // verificao de erro no comando de inicio da aquisio
   if (rstart.PacketCode == pcAcqStartR && rstart.Error == 0)
     printf("\nUDPdrv: Inicializao das Aquisies OK");
   else
     printf("\nUDPdrv: Erro na Inicializao das Aquisies");
}

static void mdlInitializeSampleTimes(SimStruct *S)
{
   ssSetSampleTime(S, 0, CONTINUOUS_SAMPLE_TIME);
   ssSetOffsetTime(S, 0, FIXED_IN_MINOR_STEP_OFFSET);
}

/* Tarefas do loop Scan */
static void mdlOutputs(SimStruct *S, int_T tid)
{
   int i, j; // contadores
   InputRealPtrsType AnaInpPtrs = ssGetInputPortRealSignalPtrs(S,0); // Ponteiro entradas analgicas do bloco (sada analgica do equipamento)
   InputRealPtrsType DigInpPtrs = ssGetInputPortRealSignalPtrs(S,1); // Ponteiro entradas digitais do bloco (sada digital do equipamento)
   real_T *AnaOut  = ssGetOutputPortRealSignal(S,0); // Ponteiro sadas analgicas do bloco (entrada analgica do equipamento)
   real_T *DigOut  = ssGetOutputPortRealSignal(S,1); // Ponteiro sadas digitais do bloco (entrada digital do equipamento)
   real_T *AcqStat = ssGetOutputPortRealSignal(S,2); // Ponteiro sada de status do bloco

   // Verifica se  primeiro Scan, caso sim ajusta base de temporizao
   if (IsFirstStep == TRUE) {
     basetsclast=gettsc();
     IsFirstStep = FALSE;
   }
   
   //Consome pacote do buffer de entrada 
   if (recvfrom(ListenSocket, (char *)&dados, sizeof(dados), 0,(SOCKADDR *)&addr_recv, &addr_recv_len) == -1);

   for (i=0;i<16;i++){
     aout.Value =  *AnaInpPtrs[i]*65535/20; // Converte valor de entrada do bloco para escala (-10 V a 10 V) do equipamento 
     aout.AddrADS = (16+i) << 8; // end. dos canais analogicos comeando em 16 no Byte + significativo 
     // envia comando de escrita na saida analgica
     if (sendto(ListenSocket, (char *)&aout, sizeof(aout), 0,(SOCKADDR *)&addr_send, addr_send_len) == -1)
        printf("\nUDPdrv: Impossvel gravar valores na sada analgica(%d)", i+1);
     waitbusy(0.4e-3, gettsc(), CalFreq);// tempo de espera p/ equipamento reconhecer comandos de escrita
   }

   //Atualiza as Sadas Digitais
   dout.Value = 0;
   dout.AddrADS = 0;
   for (i=0;i<16;i++)
     if (*DigInpPtrs[i]!= 0.0)
         dout.Value = dout.Value + (1<<i); //L valores da entradas digitais do bloco e converte para um valor de 16 bits   }
       //Atualiza as Sadas Analgicas

   // envia comando de escrita na saida digital
   if (sendto(ListenSocket, (char *)&dout, sizeof(dout), 0,(SOCKADDR *)&addr_send, addr_send_len) == -1)
     printf("\nUDPdrv: Impossvel gravar valores na sada digital");
 
   //Realiza a leitura do pacote com dados da amostragem    
   if (recvfrom(ListenSocket, (char *)&dados, sizeof(dados),0,(SOCKADDR *)&addr_recv, &addr_recv_len) == -1);
   
   if (dados.PacketCode == pcAcqData){ // Verifica se o pacote  de dados
     // Atualiza as sadas analgicas do bloco (entradas analgica do equipamento)
     for (i=0;i<16;i++)
       AnaOut[i] = (real_T) dados.Sample[i]*escala; // Ajusta valor conforme o range do bloco
     // Atualiza as sadas digitais do bloco (entradas digitais do equipamento)
     for (i=0;i<16;i++){
       if ((dados.Sample[16] & 0x0001) == 1) // Compara timo bit dos canais digitais com 1
         DigOut[i]= 1;                       // Se for igual a "1" o valor da sada digital correpondente  "1"
       else
         DigOut[i]= 0;                       // Se for igual a "0" o valor da sada digital correpondente  "0"
       dados.Sample[16] = dados.Sample[16] >> 1; // Desloca 1 bit para direita 
     }
     AcqStat[2]=0;// pacote de dados lido
   }
   else {
     errodados++; // Aumenta contador de erro de recepo de dados;
     AcqStat[2]=1;// no era um pacote de dados 
   }

   // temporizao
   AcqStat[1] = timelapse(basetscfirst, CalFreq); // Tempo decorrido
   remainingtime = waitbusy(1/WorkFreq, basetsclast, CalFreq); 
   AcqStat[0] = (double) dados.FrameNo; // Nmero do pacote
   basetsclast=gettsc();
}

/* Finalizao */
static void mdlTerminate(SimStruct *S)
{
   int i,j; // Contadores

   // Para a aquisio de dados
   stop.PacketCode = pcAcqStop;
   stop.Request = 0x0000;
   stop.Reserved = 0x0000;
   // Envia comando para parar aquisio
   if (sendto(ListenSocket, (char *)&stop, sizeof(stop), 0,(SOCKADDR *)&addr_send, addr_send_len) == -1)
     printf("\nUDPdrv: Erro no envio do comando de stop");
   // Aguarda resposta do comando de aquisio
   
   waitbusy(1e-3, gettsc(), CalFreq);// tempo de espera p/ equipamento reconhecer comandos 
   // Consome buffer
   if (recvfrom(ListenSocket, (char *)&rstop, sizeof(rstop), 0,(SOCKADDR *)&addr_recv, &addr_recv_len) == -1);
   
   waitbusy(1e-3, gettsc(), CalFreq);// tempo de espera p/ equipamento reconhecer comandos 
   // Resposta do comando de para a aquisio
   if (recvfrom(ListenSocket, (char *)&rstop, sizeof(rstop), 0,(SOCKADDR *)&addr_recv, &addr_recv_len) == -1)
     printf("\nUDPdrv: Erro na resposta do stop");

   // Verificao de erro no comando de parar a aquisio
   if (rstop.PacketCode == pcAcqStopR  && rstop.Error == 0)
     printf("\nUDPdrv: Fim das Aquisies OK");
   else
     printf("\nUDPdrv: Erro no Fim das Aquisies");
  
   //Coloca em 0 V as 16 sadas analgicas
   aout.PacketCode = pcWriteDO;
   aout.Reserved = 0;
   aout.AddrADS = 0; 
   aout.Value = 0x0000;
   for (i=0;i<16;i++){
     aout.AddrADS = (16+i) << 8; // end. dos canais analogicos comeando em 16 no Byte + significativo 
     // envia comando de escrita na saida
     if (sendto(ListenSocket, (char *)&aout, sizeof(aout), 0,(SOCKADDR *)&addr_send, addr_send_len) == -1)
        printf("\nUDPdrv: Impossvel gravar valores na sada analgica(%d)", i+1);
     waitbusy(0.4e-3, gettsc(), CalFreq);// tempo de espera p/ equipamento reconhecer comandos de escrita
   }

   //Coloca em "0" as sadas digitais
   dout.PacketCode = pcWriteDO;
   dout.Reserved = 0;
   dout.AddrADS = 0;
   dout.Value = 0x0000;
   // envia comando de escrita na saida digital
   if (sendto(ListenSocket, (char *)&dout, sizeof(dout), 0,(SOCKADDR *)&addr_send, addr_send_len) == -1)
     printf("\nUDPdrv: Impossvel gravar valores na sada digital");
   
   // Motra nmero pacotes de dados com erro
   printf("\nUDPdrv: Erro em %d pacote(s)\n",errodados);

   // Fecha socket e biblioteca Winsock2
   closesocket(ListenSocket);
   WSACleanup();
}

/* Simulink stuff */
#ifdef  MATLAB_MEX_FILE    /* Is this file being compiled as a MEX-file? */
#include "simulink.c"      /* MEX-file interface mechanism */
#else
#include "cg_sfun.h"       /* Code generation registration function */
#endif
