// TSCaccess.c
// Real-time sync auxiliary routines
//
// V1.0 may/2007
// RP Marques
//
// Uses copyrighted material (see comments)

#include "windows.h"

// TSC access function (see Copyright Note in the end of file)
__forceinline ULONGLONG gettsc(void) {
 ULONG hi,lo;
 LARGE_INTEGER r;
 __asm {
     rdtsc
     mov hi,edx
     mov lo,eax
 }
 r.HighPart = hi;
 r.LowPart = lo;
 return r.QuadPart;
}

// Get time lapse in seconds
__forceinline double timelapse(ULONGLONG basetsc, double calfreq) {
   ULONGLONG currenttsc;
   currenttsc = gettsc();
   return ( currenttsc-basetsc )/calfreq;
}

// Wait for the real time to pair with simulation time
__forceinline double waitbusy(double finaltime, ULONGLONG basetsc, double calfreq) {
   double remainingtime;
   remainingtime = finaltime - timelapse(basetsc,calfreq);
   do {
       remainingtime = finaltime - timelapse(basetsc,calfreq);
   } while (remainingtime > 0);
   return remainingtime;
}

// Notice on use of copyrighted material

// Copyright (c) 2003, Keith Wansbrough
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
//     * Redistributions of source code must retain the above
//       copyright notice, this list of conditions and the following
//       disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of the University of Cambridge nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//

// (for information only: this is the BSD license as found on
//  http://www.opensource.org/licenses/bsd-license.php)
